import trimesh, numpy as np
from numpy import radians, sin, cos, sqrt, pi

from scipy.spatial import KDTree

from welleng.utils import HLA_to_NEV, get_sigmas

class WellMesh:

    def __init__(
        self,
        survey,
        n_verts=12,
        sigma=3.0,
        sigma_pa=0.5,
        Sm=0,
        method="mesh_ellipse",
    ):
        """
        Create a WellMesh object from a welleng Survey object and a
        welleng Clearance object.

        Parameters
        ----------
            survey: welleng.Survey
            clearance: welleng.Clearance
            n_verts: int
                The number of vertices along the uncertainty ellipse
                edge from which to construct the mesh.
            method: str (default="ellipse")
                The method for constructing the uncertainty edge.
                Either "ellipse" or "pedal_curve".
        """
        self.s = survey
        # self.c = clearance
        self.n_verts = int(n_verts)
        self.sigma = sigma
        self.radius = self.s.radius
        self.Sm = Sm
        self.sigma_pa = sigma_pa

        assert method in ["mesh_ellipse", "mesh_pedal_curve"], "Invalid method (ellipse or pedal_curve)"
        self.method = method

        self.sigmaH, self.sigmaL, self.sigmaA = get_sigmas(self.s.cov_hla)
        self.nevs = np.array([self.s.n, self.s.e, self.s.tvd]).T
        self._get_vertices()
        self._align_verts()

        self._get_faces()
        self._make_trimesh()       

    ### Helper functions ###

    def _get_faces(self):
        step = self.n_verts
        faces = []
        total_verts = len(self.vertices.reshape(-1,3))
        rows = int(total_verts / self.n_verts)
        
        # make first end        
        B = np.arange(1, step - 1, 1)
        C = np.arange(2, step, 1)
        A = np.zeros_like(B)
        temp = np.array([A, B, C]).T
        faces.extend(temp.tolist())
            
        # make cylinder
        temp = [np.array([step, 0, step - 1, 2 * step - 1, step, step - 1])]
        for row in range(0, rows - 1):
            verts = row * self.n_verts

            A_start = verts
            B_start = verts + step
            C_start = verts + step + 1
            D_start = verts + 1
            
            A_stop = verts + step - 1
            B_stop = verts + step + step - 1
            C_stop = verts + step + step
            D_stop = verts + step

            A = np.arange(A_start, A_stop)
            B = np.arange(B_start, B_stop)
            C = np.arange(C_start, C_stop)
            D = np.arange(D_start, D_stop)

            temp.extend(np.array([C, D, A, B, C, A]).T)

            last = np.array([B_start, A_start, D_stop - 1, C_stop - 1, B_start, D_stop -1])

            temp.extend([last])
        
        faces.extend(np.stack(temp, axis=0).reshape(-1,3).tolist())
            
        # make final end     
        B = np.arange(total_verts - step + 1, (total_verts - 1), 1)
        C = np.arange(total_verts - step + 2, (total_verts), 1)
        A = np.full_like(B, total_verts - step)
        temp = np.array([A, B, C]).T
        faces.extend(temp.tolist())
        self.faces = faces

    def _get_vertices(
        self,
        ):

        h = (np.array(self.sigmaH) * self.sigma + self.radius + self.Sm + self.sigma_pa / 2).reshape(-1,1)
        l = (np.array(self.sigmaL) * self.sigma + self.radius + self.Sm + self.sigma_pa / 2).reshape(-1,1)
        # a = self.s.sigmaA * self.c.k + self.s.radius + self.c.Sm

        if self.method == "mesh_ellipse":
            lam = np.linspace(0, 2 * pi, self.n_verts, endpoint=False)
            # theta = np.zeros_like(lam)

            x = h * cos(lam)
            y = l * sin(lam)
            z = np.zeros_like(x)
            vertices = np.stack((x, y, z), axis=-1)

        else:
            lam = np.linspace(0, 2 * pi, self.n_verts, endpoint=False)
            f = h * (
                (l ** 2 * np.cos(lam))
                / 
                (l ** 2 * (np.cos(lam)) ** 2 + h ** 2 * (np.sin(lam)) ** 2)
            )

            g = l * (
                (h ** 2 * np.sin(lam))
                /
                (l ** 2 * (np.cos(lam)) ** 2 + h ** 2 * (np.sin(lam)) ** 2)
            )
            z = np.zeros_like(f)
            vertices = np.stack((f, g, z), axis=-1)
            
        
        vertices = HLA_to_NEV(self.s.survey_rad, vertices, cov=False)
        
        self.vertices = np.vstack([
            v + nev for v, nev in zip(vertices, self.nevs)
        ]).reshape(vertices.shape)


    def _get_vertices_vectors(self, vertices, pos_center):
        vectors = vertices - pos_center
        normals = np.linalg.norm(vectors, axis=-1)
        with np.errstate(divide='ignore', invalid='ignore'):
            unit_vectors = vectors / normals.reshape(-1,1)
        unit_vectors[np.where(normals == 0)] = vectors[np.where(normals == 0)]
    
        return (vectors, normals, unit_vectors)


    def _align_verts(self):
        """
        The meshing gets unstable when the inclination is close to vertical (since high side
        can quickly shift azimuth). This code cycles through the ellipses and makes sure that
        they're all lined up relative to each other.

        """
        verts_new_list = []
        for i, verts in enumerate(self.vertices):
            if i == 0:
                verts_new_list.append([verts])
                continue
            else:
                vecs_1, norms_1, units_1 = self._get_vertices_vectors(verts_new_list[-1][0], self.nevs[i - 1])
                vecs_2, norms_2, units_2 = self._get_vertices_vectors(self.vertices[i], self.nevs[i])

                nearest = KDTree(units_2)

                key = units_1[0]
                idx_key = nearest.query(key)[1]

                rotation = units_1[3]
                idx_rotation = nearest.query(rotation)[1]

                # re-order vecs_2 to align with vecs_1
                verts_new = np.vstack((verts[idx_key:], verts[:idx_key]))

                if (idx_rotation - idx_key) < 0 and abs(idx_rotation - idx_key) < len(verts) / 2:
                    verts_new = verts_new[::-1]

                verts_new_list.append([verts_new])
            
        self.vertices = np.vstack(verts_new_list).reshape(self.vertices.shape)

    def _make_trimesh(self):
        self.mesh = trimesh.Trimesh(
            vertices=self.vertices.reshape(-1,3),
            faces=self.faces,
            process=True
        )

        if len(trimesh.repair.broken_faces(self.mesh)) > 0:
            trimesh.repair.fill_holes(self.mesh)

def make_trimesh_scene(data):
    """
    Construct a trimesh scene. A collision manager can't be saved, but a scene can and a scene can be imported
    into a collision manager.
    """
    scene = trimesh.scene.scene.Scene()
    for well in data:
        mesh = data[well].mesh
        scene.add_geometry(mesh, node_name=well, geom_name=well, parent_node_name=None)
    
    return scene

def transform_trimesh_scene(scene, origin=None, scale=100, redux=0.25):
    """
    Transforms a scene by scaling it, reseting the origin/datum and performing
    a reduction in the number of triangles to reduce the file size.

    Params:
        scene: trimesh.scene
            A trimesh scene of well meshes
        origin: 3d array [x, y, z]
            The origin of the scene from which the new scene will reset to [0, 0, 0]
        scale: float
            A scalar reduction will be performed using this float
        redux: float
            The desired reduction ratio for the number of triangles in each mesh

    """
    i = 0
    if not origin:
        T = np.array([0, 0, 0])
    else:
        T = np.array(origin)
    scene_transformed = trimesh.scene.scene.Scene()
    for well in scene.geometry.items():
        name, mesh = well
        mesh_new = mesh.copy()

        mesh_new.simplify_quadratic_decimation(int(len(mesh_new.triangles) * redux))
        mesh_new.vertices -= T

        ### change axis convention for visualisation ###
        x, y, z = mesh_new.vertices.T
        mesh_new.vertices = (np.array([x, z * -1, y]) / scale).T
        scene_transformed.add_geometry(mesh_new, node_name=name, geom_name=name)

    return scene_transformed


import numpy as np
from numpy import sin, cos, tan, pi, sqrt

import welleng.error
from welleng.utils import NEV_to_HLA

class iscwsaMwd:
    def __init__(
        self,
        error,
    ):
        """
        Class using the ISCWSA MWD (Rev4) model to determine
        well bore uncertatinty.
        """
        error.__init__
        self.e = error

        self.errors = {
                "DRFR": self._DRFR(self.e.survey, mag=self.e.errors_mag['DRFR']),
                "DSFS": self._DSFS(self.e.survey, mag=self.e.errors_mag['DSFS']),
                "DSTG": self._DSTG(self.e.survey, self.e.TVD, mag=self.e.errors_mag['DSTG']),
                "ABXY_TI1S": self._ABXY_TI1S(self.e.survey, self.e.AziMag, self.e.G, self.e.Dip, mag=self.e.errors_mag['AB']),
                "ABXY_TI2S": self._ABXY_TI2S(self.e.survey, self.e.AziMag, self.e.G, self.e.Dip, mag=self.e.errors_mag['AB']),
                "ABZ": self._ABZ(self.e.survey, self.e.AziMag, self.e.G, self.e.Dip, mag=self.e.errors_mag['AB']),
                "ASXY_TI1S": self._ASXY_TI1S(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['AS']),
                "_ASXY_TI2S": self._ASXY_TI2S(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['AS']),
                "ASXY_TI3S": self._ASXY_TI3S(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['AS']),
                "ASZ": self._ASZ(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['AS']),
                "MBXY_TI1S": self._MBXY_TI1S(self.e.survey, self.e.AziMag, self.e.Dip, self.e.BTotal, mag=self.e.errors_mag['MB']),
                "MBXY_TI2S": self._MBXY_TI2S(self.e.survey, self.e.AziMag, self.e.Dip, self.e.BTotal, mag=self.e.errors_mag['MB']),
                "MBZ": self._MBZ(self.e.survey, self.e.AziMag, self.e.Dip, self.e.BTotal, mag=self.e.errors_mag['MB']),
                "MSXY_TI1S": self._MSXY_TI1S(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['MS']),
                "MSXY_TI2S": self._MSXY_TI2S(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['MS']),
                "MSXY_TI3S": self._MSXY_TI3S(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['MS']),
                "MSZ": self._MSZ(self.e.survey, self.e.AziMag, self.e.Dip, mag=self.e.errors_mag['MS']),
                "DECG": self._DECG(self.e.survey, mag=self.e.errors_mag['DECG']),
                "DECR": self._DECR(self.e.survey, mag=self.e.errors_mag['DECR']),
                "DBHG": self._DBHG(self.e.survey, self.e.Dip, self.e.BTotal, mag=self.e.errors_mag['DBHG']),
                "DBHR": self._DBHR(self.e.survey, self.e.Dip, self.e.BTotal, mag=self.e.errors_mag['DBHR']),
                "AMIL": self._AMIL(self.e.survey, self.e.AziMag, self.e.Dip, self.e.BTotal, mag=self.e.errors_mag['AMIL']),
                "SAG": self._SAG(self.e.survey, mag=self.e.errors_mag['SAG']),
                "XYM1": self._XYM1(self.e.survey, mag=self.e.errors_mag['XYM']),
                "XYM2": self._XYM2(self.e.survey, mag=self.e.errors_mag['XYM']),
                "XYM3": self._XYM3(self.e.survey, self.e.AziTrue, mag=self.e.errors_mag['XYM']),
                "XYM4": self._XYM4(self.e.survey, self.e.AziTrue, mag=self.e.errors_mag['XYM']),            
            }

        self.cov_NEVs = np.zeros((3,3,len(self.e.survey)))
        for _, value in self.errors.items():
            self.cov_NEVs += value.cov_NEV

        self.cov_HLAs = NEV_to_HLA(self.e.survey, self.cov_NEVs)

    ### error functions ###
    def _DRFR(self, survey, mag=0.35, propagation='random', NEV=True):
        dpde = np.full((len(survey), 3), [1,0,0])
        e_DIA = dpde * mag
        
        return self.e._generate_error('DRFR', e_DIA, propagation, NEV)

    def _DSFS(self, survey, mag=0.00056, propagation='systematic', NEV=True):
        dpde = np.full((len(survey), 3), [1,0,0])
        dpde = dpde * np.array(survey)
        e_DIA = dpde * mag

        return self.e._generate_error('DSFS', e_DIA, propagation, NEV)

    def _DSTG(self, survey, TVD, mag=0.00000025, propagation='systematic', NEV=True):
        dpde = np.full((len(survey), 3), [1,0,0])
        dpde[:,0] = TVD
        dpde = dpde * np.array(survey)
        e_DIA = dpde * mag

        return self.e._generate_error('DSTG', e_DIA, propagation, NEV)

    def _ABXY_TI1S(self, survey, AziMag, gfield, dip, mag=0.0040, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = -cos(survey[:,1]) / gfield
        dpde[:,2] = (cos(survey[:,1]) * tan(dip) * sin(AziMag)) / gfield
        e_DIA = dpde * mag

        return self.e._generate_error('ABXY_TI1S', e_DIA, propagation, NEV)

    def _ABXY_TI2S(self, survey, AziMag, gfield, dip, mag=0.004, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        with np.errstate(divide='ignore', invalid='ignore'):
            dpde[:,2] = np.nan_to_num(((tan(-(survey[:,1]) + (pi/2)) - tan(dip) * cos(AziMag)) / gfield),
                                posinf = 0.0,
                                neginf = 0.0
                                )
        e_DIA = dpde * mag

        sing = np.where(survey[:,1] < self.e.VerticalIncLimit)
        if len(sing[0]) < 1:
            return self.e._generate_error('ABXY_TI2S', e_DIA, propagation, NEV)
        else: 
            e_NEV = self.e._e_NEV(e_DIA)
            N = np.array(0.5 * self.e.drdp_sing['double_delta_md'] * -sin(self.e.drdp_sing['azi2']) * mag) / gfield
            E = np.array(0.5 * self.e.drdp_sing['double_delta_md'] * cos(self.e.drdp_sing['azi2']) * mag) / gfield
            V = np.zeros_like(N)
            e_NEV_sing = np.vstack((np.zeros((1,3)), np.stack((N, E, V), axis=-1), np.zeros((1,3))))
            e_NEV[sing] = e_NEV_sing[sing]

            e_NEV_star = self.e._e_NEV_star(e_DIA)
            N = np.array(0.5 * self.e.drdp_sing['delta_md'] * -sin(self.e.drdp_sing['azi2']) * mag) / gfield
            E = np.array(0.5 * self.e.drdp_sing['delta_md'] * cos(self.e.drdp_sing['azi2']) * mag) / gfield
            V = np.zeros_like(N)
            e_NEV_star_sing = np.vstack((np.zeros((1,3)), np.stack((N, E, V), axis=-1), np.zeros((1,3))))
            e_NEV_star[sing] = e_NEV_star_sing[sing]

            return self.e._generate_error('ABXY_TI2S', e_DIA, propagation, NEV, e_NEV, e_NEV_star)

    def _ABZ(self, survey, AziMag, gfield, dip, mag=0.004, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = -sin(np.array(survey)[:,1]) / gfield
        dpde[:,2] = (sin(np.array(survey)[:,1]) * tan(dip) * sin(AziMag)) / gfield
        e_DIA = dpde * mag

        return self.e._generate_error('ABZ', e_DIA, propagation, NEV)

    def _ASXY_TI1S(self, survey, AziMag, dip, mag=0.0005, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = sin(np.array(survey)[:,1]) * cos(np.array(survey)[:,1]) / sqrt(2)
        dpde[:,2] = (sin(np.array(survey)[:,1]) * -tan(dip) * cos(np.array(survey)[:,1]) * sin(AziMag)) / sqrt(2)
        e_DIA = dpde * mag

        return self.e._generate_error('ASXY_TI1S', e_DIA, propagation, NEV)

    def _ASXY_TI2S(self, survey, AziMag, dip, mag=0.0005, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = sin(np.array(survey)[:,1]) * cos(np.array(survey)[:,1]) / 2
        dpde[:,2] = (sin(np.array(survey)[:,1]) * -tan(dip) * cos(np.array(survey)[:,1]) * sin(AziMag)) / 2
        e_DIA = dpde * mag

        return self.e._generate_error('ASXY_TI2S', e_DIA, propagation, NEV)

    def _ASXY_TI3S(self, survey, AziMag, dip, mag=0.0005, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = (sin(np.array(survey)[:,1]) * tan(dip) * cos(AziMag) - cos(np.array(survey)[:,1])) / 2
        e_DIA = dpde * mag

        return self.e._generate_error('ASXY_TI3S', e_DIA, propagation, NEV)

    def _ASZ(self, survey, AziMag, dip, mag=0.0005, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = -sin(np.array(survey)[:,1]) * cos(np.array(survey)[:,1])
        dpde[:,2] = (sin(np.array(survey)[:,1]) * tan(dip) * cos(np.array(survey)[:,1]) * sin(AziMag))
        e_DIA = dpde * mag

        return self.e._generate_error('ASZ', e_DIA, propagation, NEV)

    def _MBXY_TI1S(self, survey, AziMag, dip, bfield, mag=70.0, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = (-cos(np.array(survey)[:,1]) * sin(AziMag)) / (bfield * cos(dip))
        e_DIA = dpde * mag

        return self.e._generate_error('MBXY_TI1S', e_DIA, propagation, NEV)

    def _MBXY_TI2S(self, survey, AziMag, dip, bfield, mag=70.0, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = cos(AziMag) / (bfield * cos(dip))
        e_DIA = dpde * mag

        return self.e._generate_error('MBXY_TI2S', e_DIA, propagation, NEV)

    def _MBZ(self, survey, AziMag, dip, bfield, mag=70.0, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = (-sin(np.array(survey)[:,1]) * sin(AziMag)) / (bfield * cos(dip))
        e_DIA = dpde * mag

        return self.e._generate_error('MBZ', e_DIA, propagation, NEV)

    def _MSXY_TI1S(self, survey, AziMag, dip, mag=0.0016, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = sin(np.array(survey)[:,1]) * sin(AziMag) * (tan(dip) * cos(np.array(survey)[:,1])
        + sin(np.array(survey)[:,1]) * cos(AziMag)) / sqrt(2)
        e_DIA = dpde * mag

        return self.e._generate_error('MSXY_TI1S', e_DIA, propagation, NEV)


    def _MSXY_TI2S(self, survey, AziMag, dip, mag=0.0016, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = sin(AziMag) * (tan(dip) * sin(np.array(survey)[:,1]) * cos(np.array(survey)[:,1])
        - cos(np.array(survey)[:,1]) * cos(np.array(survey)[:,1]) * cos(AziMag) -cos(AziMag)) / 2
        e_DIA = dpde * mag

        return self.e._generate_error('MSXY_TI2S', e_DIA, propagation, NEV)

    def _MSXY_TI3S(self, survey, AziMag, dip, mag=0.0016, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = (cos(np.array(survey)[:,1]) * cos(AziMag) * cos(AziMag)
                - cos(np.array(survey)[:,1]) * sin(AziMag) * sin(AziMag) 
                - tan(dip) * sin(np.array(survey)[:,1]) * cos(AziMag)) / 2
        e_DIA = dpde * mag

        return self.e._generate_error('MSXY_TI3S', e_DIA, propagation, NEV)

    def _MSZ(self, survey, AziMag, dip, mag=0.0016, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = -(sin(np.array(survey)[:,1]) * cos(AziMag)
        + tan(dip) * cos(np.array(survey)[:,1])) * sin(np.array(survey)[:,1]) * sin(AziMag)
        e_DIA = dpde * mag

        return self.e._generate_error('MSZ', e_DIA, propagation, NEV)

    def _DECG(self, survey, mag=0.00628, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = 1
        e_DIA = dpde * mag

        return self.e._generate_error('DECG', e_DIA, propagation, NEV)

    def _DECR(self, survey, mag=0.00175, propagation='random', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = 1
        e_DIA = dpde * mag

        return self.e._generate_error('DECR', e_DIA, propagation, NEV)

    def _DBHG(self, survey, dip, bfield, mag=np.radians(5000), propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = 1 / (bfield * cos(dip))
        e_DIA = dpde * mag

        return self.e._generate_error('DBHG', e_DIA, propagation, NEV)

    def _DBHR(self, survey, dip, bfield, mag=np.radians(3000), propagation='random', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = 1 / (bfield * cos(dip))
        e_DIA = dpde * mag

        return self.e._generate_error('DBHR', e_DIA, propagation, NEV)

    def _AMIL(self, survey, AziMag, dip, bfield, mag=220.0, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = -sin(np.array(survey)[:,1]) * sin(AziMag) / (bfield * cos(dip))
        e_DIA = dpde * mag

        return self.e._generate_error('AMIL', e_DIA, propagation, NEV)

    def _SAG(self, survey, mag=0.00349, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = sin(np.array(survey)[:,1])
        e_DIA = dpde * mag

        return self.e._generate_error('SAG', e_DIA, propagation, NEV)

    def _XYM1(self, survey, mag=0.00175, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = np.absolute(sin(np.array(survey)[:,1]))
        e_DIA = dpde * mag

        return self.e._generate_error('XYM1', e_DIA, propagation, NEV)

    def _XYM2(self, survey, mag=0.00175, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,2] = -1
        e_DIA = dpde * mag

        return self.e._generate_error('XYM2', e_DIA, propagation, NEV)

    def _XYM3(self, survey, AziTrue, mag=0.00175, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = np.absolute(cos(np.array(survey)[:,1])) * cos(AziTrue)
        with np.errstate(divide='ignore', invalid='ignore'):
            dpde[:,2] = np.nan_to_num(
                -(np.absolute(cos(np.array(survey)[:,1])) * sin(AziTrue)) / sin(np.array(survey)[:,1]),
                posinf = 0.0,
                neginf = 0.0
            )
        e_DIA = dpde * mag

        sing = np.where(survey[:,1] < self.e.VerticalIncLimit)
        if len(sing[0]) < 1:
            return self.e._generate_error('XYM3', e_DIA, propagation, NEV)
        else: 
            e_NEV = self.e._e_NEV(e_DIA)
            N = np.array(0.5 * self.e.drdp_sing['double_delta_md'] * mag)
            E = np.zeros(len(self.e.drdp_sing['double_delta_md']))
            V = np.zeros_like(N)
            e_NEV_sing = np.vstack((np.zeros((1,3)), np.stack((N, E, V), axis=-1), np.zeros((1,3))))
            e_NEV[sing] = e_NEV_sing[sing]

            e_NEV_star = self.e._e_NEV_star(e_DIA)
            N = np.array(0.5 * self.e.drdp_sing['delta_md'] * mag)
            E = np.zeros(len(self.e.drdp_sing['delta_md']))
            V = np.zeros_like(N)
            e_NEV_star_sing = np.vstack((np.zeros((1,3)), np.stack((N, E, V), axis=-1), np.zeros((1,3))))
            e_NEV_star[sing] = e_NEV_star_sing[sing]

            return self.e._generate_error('XYM3', e_DIA, propagation, NEV, e_NEV, e_NEV_star)


    def _XYM4(self, survey, AziTrue, mag=0.00175, propagation='systematic', NEV=True):
        dpde = np.zeros((len(survey), 3))
        dpde[:,1] = np.absolute(cos(np.array(survey)[:,1])) * sin(AziTrue)
        with np.errstate(divide='ignore', invalid='ignore'):
            dpde[:,2] = np.nan_to_num((np.absolute(np.cos(np.array(survey)[:,1])) * cos(AziTrue))
            / sin(np.array(survey)[:,1]),
                posinf = 0.0,
                neginf = 0.0
            )
        e_DIA = dpde * mag

        sing = np.where(survey[:,1] < self.e.VerticalIncLimit)
        if len(sing[0]) < 1:
            return self.e._generate_error('XYM4', e_DIA, propagation, NEV)
        else: 
            e_NEV = self.e._e_NEV(e_DIA)
            N = np.zeros(len(self.e.drdp_sing['double_delta_md']))
            E = np.array(0.5 * self.e.drdp_sing['double_delta_md'] * mag)
            V = np.zeros_like(N)
            e_NEV_sing = np.vstack((np.zeros((1,3)), np.stack((N, E, V), axis=-1), np.zeros((1,3))))
            e_NEV[sing] = e_NEV_sing[sing]

            e_NEV_star = self.e._e_NEV_star(e_DIA)
            N = np.zeros(len(self.e.drdp_sing['delta_md']))
            E = np.array(0.5 * self.e.drdp_sing['delta_md'] * mag)
            V = np.zeros_like(N)
            e_NEV_star_sing = np.vstack((np.zeros((1,3)), np.stack((N, E, V), axis=-1), np.zeros((1,3))))
            e_NEV_star[sing] = e_NEV_star_sing[sing]

            return self.e._generate_error('XYM4', e_DIA, propagation, NEV, e_NEV, e_NEV_star)


from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()

with open("VERSION") as f:
    version = f.read()
download_url = f'https://github.com/jonnymaserati/welleng/archive/v{version}.tar.gz'

setup(
    name='welleng',
    version=version,    
    description='A collection of Well Engineering tools',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/jonnymaserati/welleng',
    download_url=download_url,
    keywords=['well', 'trajectory', 'wellpath', 'wellbore', 'drilling', 'error', 'separation', 'minimum curvature', 'iscwsa'],
    author='Jonathan Corcutt',
    author_email='jonnycorcutt@gmail.com',
    license='LGPL v3',
    packages=find_packages(exclude=["tests"]),
    install_requires=required,
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',  
        'Operating System :: OS Independent',        
        'Programming Language :: Python :: 3',
    ],
    python_requires='>=3.8',
)